# -*- coding: utf-8 -*-
"""PyRGG params."""
from textwrap import dedent, fill
import os

PYRGG_VERSION = "1.7"

os.environ["PYRGG_TEST_MODE"] = "0"

MENU_ITEMS = {
    1: ["engine", dedent(
        """\
        - Select generation engine :
        1- PyRGG
        2- Erdos-Renyi-Gilbert - G(n, p)
        3- Erdos-Renyi - G(n, m)
        4- Stochastic Block Model - {C1, C2, ..., Ck}
        """
    )],
    2: ["file_name", "- File Name (Not Empty) : "],
    3: ["number_of_files", "- Number of Files (>=0) : "],
    4: ["output_format", dedent(
        """\
        - Graph Formats :
        1- DIMACS(.gr)
        2- JSON(.json)
        3- CSV(.csv)
        4- YAML(.yaml)
        5- WEL(.wel)
        6- ASP(.lp)
        7- Pickle(.p)
        8- UCINET DL(.dl)
        9- TGF(.tgf)
        10- TSV(.tsv)
        11- Matrix Market(.mtx)
        12- Graph Line(.gl)
        13- GDF(.gdf)
        14- GML(.gml)
        15- GEXF(.gexf)
        16- DOT(.gv)
        """
    )],
    5: ["config", "- Save Config[1] or Not[0]"],
}

SUFFIX_MENU = {
    1: ".gr",
    2: ".json",
    3: ".csv",
    4: ".yaml",
    5: ".wel",
    6: ".lp",
    7: ".p",
    8: ".dl",
    9: ".tgf",
    10: ".tsv",
    11: ".mtx",
    12: ".gl",
    13: ".gdf",
    14: ".gml",
    15: ".gexf",
    16: ".gv"
}

ENGINE_MENU = {
    1: "pyrgg",
    2: "erg",
    3: "er",
    4: "sbm",
}

ENGINE_MENU_INV = {v: k for k, v in ENGINE_MENU.items()}

PYRGG_ENGINE_PARAMS = {
    1: ["vertices", "- Vertices Number (n >= 0) : "],
    2: ["min_edge", "- Min Edge Number - Connected to Each Vertex (>=0) : "],
    3: ["max_edge", "- Max Edge Number - Connected to Each Vertex (>=0) : "],
    4: ["weight", "- Unweighted[0] or Weighted[1]"],
    5: ["min_weight", "- Min Weight : "],
    6: ["max_weight", "- Max Weight : "],
    7: ["sign", "- Unsigned[0] or Signed[1]"],
    8: ["direct", "- Undirected[0] or Directed[1]"],
    9: ["self_loop", "- No Self Loop[0] or Self Loop[1]"],
    10: ["multigraph", "- Simple[0] or Multigraph[1]"],
}

ERG_ENGINE_PARAMS = {
    1: ["vertices", "- Vertices Number (n >= 0) : "],
    2: ["probability", "- Probability (0 <= p <= 1) : "],
    3: ["direct", "- Undirected[0] or Directed[1]"],
}

ER_ENGINE_PARAMS = {
    1: ["vertices", "- Vertices Number (n >= 0) : "],
    2: ["edge_number", "- Edge Number (m >= 0) : "],
    3: ["direct", "- Undirected[0] or Directed[1]"],
}

SBM_ENGINE_PARAMS = {
    1: ["vertices", "- Vertices Number (n >= 0) : "],
    2: ["blocks", "- Blocks Number (k >= 0) : "],
    3: ["intra_probability", "- Intra Block Probability (0 <= p_intra <= 1): "],
    4: ["inter_probability", "- Inter Block Probability (0 <= p_inter <= 1): "],
    5: ["direct", "- Undirected[0] or Directed[1]"],
    6: ["self_loop", "- No Self Loop[0] or Self Loop[1]"],
}

ENGINE_PARAM_MAP = {
    1: PYRGG_ENGINE_PARAMS,
    2: ERG_ENGINE_PARAMS,
    3: ER_ENGINE_PARAMS,
    4: SBM_ENGINE_PARAMS,
}

OUTPUT_FORMAT = {i: output_format[1:].upper()
                 for i, output_format in SUFFIX_MENU.items()}

OUTPUT_FORMAT_INV = {v: k for k, v in OUTPUT_FORMAT.items()}


SOURCE_DIR = os.getcwd()

CONFIG_FILE_FORMAT = "{}.pyrgg.config.json"

PYRGG_LINKS = """
Webpage : https://www.pyrgg.site
Repository : https://github.com/sepandhaghighi/pyrgg
Paper : https://doi.org/10.21105/joss.00331
* If you use PyRGG in your research, please cite our paper
"""

_description = """\
PyRGG is a user-friendly synthetic random graph generator that is written in Python and supports multiple graph file formats, such as DIMACS-Graph files.
It can generate graphs of various sizes and is specifically designed to create input files for a wide range of graph-based research applications, including testing,
benchmarking, and performance analysis of graph processing frameworks.
PyRGG is aimed at computer scientists who are studying graph algorithms and graph processing frameworks.
"""

PYRGG_DESCRIPTION = fill(_description, width=113)

PYRGG_INPUT_ERROR_MESSAGE = "[Error] Bad input!"

PYRGG_FILE_ERROR_MESSAGE = "[Error] Bad input file!"

PYRGG_INVALID_ENGINE_ERROR_MESSAGE = "[Error] Invalid engine!"

PYRGG_YAML_ERROR_MESSAGE = "[Error] Failed to generate YAML file!"

PYRGG_PICKLE_ERROR_MESSAGE = "[Error] Failed to generate Pickle file!"

PYRGG_LOGGER_ERROR_MESSAGE = "[Error] Logger failed!"

PYRGG_CONFIG_LOAD_ERROR_MESSAGE = "[Error] Failed to load config file!"

PYRGG_CONFIG_SAVE_ERROR_MESSAGE = "[Error] Failed to save config file!"

PYRGG_UNDIVISIBLE_WARNING_MESSAGE = "[Warning] Vertices are not divisible by blocks. The last block will have the remaining vertices."

PYRGG_SBM_WARNING_MESSAGE = "[Warning] Stochastic Block Model gets the number of blocks and inter/intra probabilities. To get more detailed configuration, please save and edit the config file."

PYRGG_CONFIG_LIST_MESSAGE = "Config files detected in the current directory are listed below:"

PYRGG_CONFIG_LOAD_MESSAGE = "Press the config index to load or any other keys to start with the new one:"

DIMACS_FIX = dedent(
    """\
    c FILE                  :{0}.gr
    c No. of vertices       :{1}
    c No. of edges          :{2}
    c Max. weight           :{3}
    c Min. weight           :{4}
    p sp {1} {2}
    """
)



# -*- coding: utf-8 -*-
"""Pyrgg functions module."""
import os
from random import randint
from json import loads as json_loads
from json import dump as json_dump
from pickle import dump as pickle_dump
from yaml import safe_dump as yaml_dump
import datetime
import pyrgg.params


def is_weighted(max_weight, min_weight, signed):
    """
    Check the graph is weighted or not.

    :param max_weight: maximum weight
    :type max_weight: int
    :param min_weight: minimum weight
    :type min_weight: int
    :param signed: weight sign flag
    :type signed: bool
    :return: result as bool
    """
    if max_weight == min_weight and min_weight == 0:
        return False
    if max_weight == min_weight and min_weight == 1 and not signed:
        return False
    return True


def get_min_max_weight(weight_dic):
    """
    Get minimum and maximum weight values.

    :param weight_dic: weight dictionary
    :type weight_dic: dict
    :return: minimum and maximum weight values
    """
    all_weights = [abs(w) for weights in weight_dic.values() for w in weights]
    return min(all_weights), max(all_weights)


def is_signed(weight_dic): # pragma: no cover
    """
    Check if the graph is signed.

    :param weight_dic: weight dictionary
    :type weight_dic: dict
    :return: signed flag
    """
    return any([any([w < 0 for w in weights]) for weights in weight_dic.values()])


def has_self_loop(edge_dic): # pragma: no cover
    """
    Check if the graph has self loops.
    
    :param edge_dic: edge dictionary
    :type edge_dic: dict
    :return: self looped flag
    """
    return any([v in edges for v, edges in edge_dic.items()])


def is_multigraph(edge_dic):
    """
    Check if the graph is a multigraph.

    :param edge_dic: edge dictionary
    :type edge_dic: dict
    :return: multigraph flag
    """
    return any([len(set(edges)) != len(edges) for edges in edge_dic.values()])


def get_precision(input_number):
    """
    Return precision of input number.

    :param input_number: input number
    :type input_number: float
    :return: precision as int
    """
    try:
        number_str = str(input_number)
        _, decimal_part = number_str.split(".")
        return len(decimal_part)
    except Exception:
        return 0


def threshold_calc(min_edge, max_edge, vertex_degree):
    """
    Calculate threshold for branch_gen_pyrgg function.

    :param min_edge: minimum number of edges (connected to each vertex)
    :type min_edge: int
    :param max_edge: maximum number of edges (connected to each vertex)
    :type max_edge: int
    :param vertex_degree: vertex degree
    :type vertex_degree: int
    :return: threshold as int
    """
    threshold = min_edge
    lower_limit = 0
    upper_limit = max_edge - vertex_degree
    if vertex_degree < min_edge:
        lower_limit = min_edge - vertex_degree
    if upper_limit > lower_limit:
        threshold = randint(lower_limit, upper_limit)
    return threshold


def is_float(input_number):
    """
    Check input for float conversion.

    :param input_number: input number
    :type input_number: float or int or str
    :return: result as bool
    """
    try:
        _, decimal_part = divmod(float(input_number), 1)
    except TypeError:
        return False
    else:
        return True if decimal_part else False


def handle_string(string):
    """
    Handle string and raise ValueError if it is empty.

    :param string: input string
    :type string: str
    :return: result as str
    """
    if string == "":
        raise ValueError
    return string


def handle_pos_int(input_number):
    """
    Handle input number and raise ValueError if it is negative.

    :param input_number: input number
    :type input_number: float or int or str
    :return: result as int
    """
    val = int(input_number)
    if val < 0:
        raise ValueError
    return val


def handle_natural_number(input_number):
    """
    Handle input number and raise ValueError if it is not a natural number.

    :param input_number: input number
    :type input_number: float or int or str
    :return: result as int
    """
    val = int(input_number)
    if val < 1:
        raise ValueError
    return val

def handle_str_to_number(string):
    """
    Convert string to float or int.

    :param string: input string
    :type string: str
    :return: result as float or int
    """
    return float(string) if is_float(string) else int(string)


def handle_str_prob(string):
    """
    Convert string to float and raise ValueError if string is invalid.

    :param string: input string
    :type string: str
    :return: result as float
    """
    val = handle_str_to_number(string)
    if val < 0:
        raise ValueError
    if val > 1:
        raise ValueError
    return val


def handle_str_to_bool(string):
    """
    Convert 0/1 string to bool and raise ValueError if string is invalid.

    :param string: input string
    :type string: str
    :return: result as bool
    """
    val = int(string)
    if val not in [0, 1]:
        raise ValueError
    return bool(val)


def handle_output_format(string):
    """
    Convert string to output format index.

    :param string: input string
    :type string: str
    :return: output format index as int
    """
    output_format = handle_pos_int(string)
    if output_format not in pyrgg.params.SUFFIX_MENU:
        raise ValueError
    return output_format


def handle_engine(string):
    """
    Convert string to engine index.

    :param string: input string
    :type string: str
    :return: engine index as int
    """
    engine = handle_pos_int(string)
    if engine not in pyrgg.params.ENGINE_MENU:
        raise ValueError
    return engine


def handle_pos_list(string):
    """
    Convert string to positive integer list.

    :param string: input string
    :type string: str
    :return: result as list
    """
    return [handle_pos_int(i) for i in string.split()]


def handle_prob_matrix(string):
    """
    Convert string to probability matrix.

    :param string: input string
    :type string: str
    :return: result as list of lists
    """
    matrix = []
    for row in string.split(","):
        matrix.append([handle_str_prob(i) for i in row.split()])
    return matrix


ITEM_HANDLERS = {
    "file_name": handle_string,
    "output_format": handle_output_format,
    "weight": handle_str_to_bool,
    "engine": handle_engine,
    "vertices": handle_pos_int,
    "number_of_files": handle_pos_int,
    "max_weight": handle_str_to_number,
    "min_weight": handle_str_to_number,
    "min_edge": handle_pos_int,
    "max_edge": handle_pos_int,
    "edge_number": handle_pos_int,
    "sign": handle_str_to_bool,
    "direct": handle_str_to_bool,
    "self_loop": handle_str_to_bool,
    "multigraph": handle_str_to_bool,
    "config": handle_str_to_bool,
    "probability": handle_str_prob,
    "blocks": handle_natural_number,
    "inter_probability": handle_str_prob,
    "intra_probability": handle_str_prob,
}


def description_print():
    """
    Print justified description for overview in console.

    :return: None
    """
    print(pyrgg.params.PYRGG_LINKS)
    line(40)
    print("\n")
    print(pyrgg.params.PYRGG_DESCRIPTION)
    print("\n")
    line(40)


def line(num=11, char="#"):
    """
    Print line of char.

    :param num: number of character in this line
    :type num : int
    :param char: character
    :type char: str
    :return: None
    """
    print(char * num)


def convert_bytes(num):
    """
    Convert num to idiomatic byte unit.

    :param num: the input number.
    :type num: int
    :return: result as str
    """
    for x in ['bytes', 'KB', 'MB', 'GB', 'TB']:
        if num < 1024.0:
            return "%3.1f %s" % (num, x)
        num /= 1024.0


def filesize(fileaddr):  # pragma: no cover
    """
    Calculate output file size.

    :param fileaddr: file addresses
    :type fileaddr: str
    :return: file size for print as string
    """
    file_info = os.stat(fileaddr)
    file_size = file_info.st_size
    print("Graph File Size : " + convert_bytes(file_size))


def time_convert(input_time):
    """
    Convert input_time from sec to DD,HH,MM,SS format.

    :param input_time: input time in sec
    :type input_time: float
    :return: converted time as str
    """
    postfix_dict = {"s": "second", "d": "day", "h": "hour", "m": "minute"}
    value_dict = {"s": 0, "d": 0, "h": 0, "m": 0}
    value_dict["s"] = float(input_time)
    value_dict["d"], value_dict["s"] = divmod(value_dict["s"], 24 * 3600)
    value_dict["h"], value_dict["s"] = divmod(value_dict["s"], 3600)
    value_dict["m"], value_dict["s"] = divmod(value_dict["s"], 60)
    for i in postfix_dict:
        if value_dict[i] != 1:
            postfix_dict[i] += "s"
    return ", ".join([
        "{0:02.0f} {1}".format(value_dict["d"], postfix_dict["d"]),
        "{0:02.0f} {1}".format(value_dict["h"], postfix_dict["h"]),
        "{0:02.0f} {1}".format(value_dict["m"], postfix_dict["m"]),
        "{0:02.0f} {1}".format(value_dict["s"], postfix_dict["s"]),
    ])


def input_filter(input_dict):
    """
    Filter input data.

    :param input_dict: input dictionary
    :type input_dict: dict
    :return: filtered data as dict
    """
    filtered_dict = input_dict.copy()
    edge_upper_threshold = filtered_dict["vertices"]
    for key in ["min_edge", "max_edge", "vertices"]:
        if filtered_dict[key] < 0:
            filtered_dict[key] *= -1

    if filtered_dict["min_weight"] > filtered_dict["max_weight"]:
        filtered_dict["min_weight"], filtered_dict["max_weight"] = (
            filtered_dict["max_weight"], filtered_dict["min_weight"]
        )

    if filtered_dict["min_edge"] > filtered_dict["max_edge"]:
        filtered_dict["min_edge"], filtered_dict["max_edge"] = (
            filtered_dict["max_edge"], filtered_dict["min_edge"]
        )

    if not filtered_dict["self_loop"]:
        edge_upper_threshold -= 1

    if not filtered_dict["multigraph"]:
        for key in ["min_edge", "max_edge"]:
            filtered_dict[key] = min(filtered_dict[key], edge_upper_threshold)

    return filtered_dict


def get_input(input_func=input):
    """
    Get input from user and return as dictionary.

    :param input_func: input function
    :type input_func: function object
    :return: inputs as dict
    """
    result_dict = {
        "file_name": "",
        "number_of_files": 1,
        "vertices": 0,
        "max_weight": 1,
        "min_weight": 1,
        "min_edge": 0,
        "max_edge": 0,
        "edge_number": 0,
        "sign": True,
        "output_format": 1,
        "weight": True,
        "engine": 1,
        "direct": True,
        "self_loop": True,
        "multigraph": False,
        "config": False,
        "probability": 0.5,
        "blocks": 1,
        "inter_probability": 0.75,
        "intra_probability": 0.25,
    }

    result_dict = _update_using_menu(result_dict, input_func)
    result_dict = _update_with_engine_params(
        result_dict, input_func, pyrgg.params.ENGINE_PARAM_MAP[result_dict['engine']])
    result_dict = _post_input_update(result_dict)
    return input_filter(result_dict)


def _update_using_menu(result_dict, input_func):
    """
    Update result_dict using user input from the menu.

    :param result_dict: result data
    :type result_dict: dict
    :param input_func: input function
    :type input_func: function object
    :return: result_dict as dict
    """
    for index in sorted(pyrgg.params.MENU_ITEMS):
        item1, item2 = pyrgg.params.MENU_ITEMS[index]
        while True:
            try:
                result_dict[item1] = ITEM_HANDLERS[item1](
                    input_func(item2)
                )
            except Exception:
                print(pyrgg.params.PYRGG_INPUT_ERROR_MESSAGE)
            else:
                break
    return result_dict


def _update_with_engine_params(result_dict, input_func, engine_params):
    """
    Update result_dict using user input based on given engine requirements.

    :param result_dict: result data
    :type result_dict: dict
    :param input_func: input function
    :type input_func: function object
    :param engine_params: engine parameters
    :type engine_params: dict
    :return: result_dict as dict
    """
    if result_dict['engine'] == 4:
        print(pyrgg.params.PYRGG_SBM_WARNING_MESSAGE)
    for index in sorted(engine_params):
        item1, item2 = engine_params[index]
        if not result_dict["weight"] and item1 in ["max_weight", "min_weight"]:
            continue
        while True:
            try:
                result_dict[item1] = ITEM_HANDLERS[item1](
                    input_func(item2)
                )
            except Exception:
                print(pyrgg.params.PYRGG_INPUT_ERROR_MESSAGE)
            else:
                break
    return result_dict


def _post_input_update(result_dict):
    """
    Update result_dict after getting user input.

    :param result_dict: result data
    :type result_dict: dict
    :return: result_dict as dict
    """
    result_dict["block_sizes"] = [result_dict["vertices"] // result_dict["blocks"]] * (result_dict["blocks"])
    if result_dict["vertices"] % result_dict["blocks"] != 0:
        result_dict["block_sizes"][-1] += result_dict["vertices"] % result_dict["blocks"]
        print(pyrgg.params.PYRGG_UNDIVISIBLE_WARNING_MESSAGE)
    result_dict["probability_matrix"] = [
        [result_dict["intra_probability"] if i == j else result_dict["inter_probability"]
            for j in range(result_dict["blocks"])]
        for i in range(result_dict["blocks"])
    ]
    return result_dict


def json_to_yaml(filename):
    """
    Convert json file to yaml file.

    :param filename: filename
    :type filename: str
    :return: None
    """
    try:
        with open(filename + ".json", "r") as json_file:
            json_data = json_loads(json_file.read())
            with open(filename + ".yaml", "w") as yaml_file:
                yaml_dump(json_data, yaml_file, default_flow_style=False)
    except FileNotFoundError:
        print(pyrgg.params.PYRGG_YAML_ERROR_MESSAGE)


def json_to_pickle(filename):
    """
    Convert json file to pickle file.

    :param filename: filename
    :type filename: str
    :return: None
    """
    try:
        with open(filename + ".json", "r") as json_file:
            json_data = json_loads(json_file.read())
            with open(filename + ".p", "wb") as pickle_file:
                pickle_dump(json_data, pickle_file)
    except FileNotFoundError:
        print(pyrgg.params.PYRGG_PICKLE_ERROR_MESSAGE)


def save_config(input_dict):
    """
    Save input_dict as the generation config.

    :param input_dict: input data
    :type input_dict: dict
    :return: path to file as str
    """
    try:
        input_dict_temp = input_dict.copy()
        input_dict_temp['engine'] = pyrgg.params.ENGINE_MENU[input_dict_temp['engine']]
        input_dict_temp['pyrgg_version'] = pyrgg.params.PYRGG_VERSION
        input_dict_temp['output_format'] = pyrgg.params.OUTPUT_FORMAT[input_dict_temp['output_format']]
        fname = pyrgg.params.CONFIG_FILE_FORMAT.format(
            input_dict_temp['file_name'])
        with open(fname, "w") as json_file:
            json_dump(input_dict_temp, json_file, indent=2)
        return os.path.abspath(fname)
    except BaseException:
        print(pyrgg.params.PYRGG_CONFIG_SAVE_ERROR_MESSAGE)


def load_config(path):
    """
    Load config based on given path.

    :param path: path to config file
    :type path: str
    :return: input data as dict
    """
    try:
        with open(path, "r") as json_file:
            config = json_loads(json_file.read())
            config['output_format'] = pyrgg.params.OUTPUT_FORMAT_INV[config['output_format']]
            config['engine'] = pyrgg.params.ENGINE_MENU_INV[config['engine']]
            return input_filter(config)
    except BaseException:
        print(pyrgg.params.PYRGG_CONFIG_LOAD_ERROR_MESSAGE)


def _print_select_config(configs, input_func=input):
    """
    Print configs in current directory and get input from user.

    :param configs: configs path
    :type configs: list
    :param input_func: input function
    :type input_func: function object
    :return: input data as dict
    """
    if len(configs) == 0:
        return None
    print(pyrgg.params.PYRGG_CONFIG_LIST_MESSAGE)
    for i, config in enumerate(configs):
        print("[{}] - {}".format(i + 1, config))
    key = input_func(pyrgg.params.PYRGG_CONFIG_LOAD_MESSAGE)
    try:
        return load_config(configs[int(key) - 1])
    except BaseException:
        return None


def check_for_config(input_func=input):
    """
    Check for config files in source directory.

    :param input_func: input function
    :type input_func: function object
    :return: input data as dict
    """
    configs = []
    for filename in os.listdir(pyrgg.params.SOURCE_DIR):
        file = os.path.join(pyrgg.params.SOURCE_DIR, filename)
        if os.path.isfile(file) and filename.endswith(
                pyrgg.params.CONFIG_FILE_FORMAT.format("")):
            configs.append(file)
    return _print_select_config(configs, input_func)


def save_log(file, file_name, elapsed_time, text):
    """
    Save generated graph logs.

    :param file: file to write log into
    :type file: file object
    :param file_name: file name
    :type file_name: str
    :param elapsed_time: elapsed time
    :type elapsed_time: str
    :param text: rest part of the text to write
    :type text: str
    :return: None
    """
    text2file = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S') + "\n"
    text2file += "Filename : {0}\n".format(file_name)
    text2file += text
    text2file += "Elapsed Time : {0}\n".format(elapsed_time)
    text2file += "-------------------------------\n"
    file.write(text2file)



# -*- coding: utf-8 -*-
"""PyRGG Engine module."""
import os
from random import randint, uniform, choice
from pyrgg.params import ENGINE_MENU, PYRGG_LOGGER_ERROR_MESSAGE
from pyrgg.functions import is_weighted, get_precision, threshold_calc
from pyrgg.functions import get_min_max_weight, is_multigraph
from pyrgg.functions import save_log


def branch_gen(
        vertex_index,
        max_edge,
        min_edge,
        min_weight,
        max_weight,
        precision,
        sign,
        direct,
        self_loop,
        multigraph,
        used_vertices,
        degree_dict,
        degree_sort_dict):
    """
    Generate branch and weight vector of each vertex.

    :param vertex_index: origin vertex index
    :type vertex_index: int
    :param max_edge: maximum number of edges (connected to each vertex)
    :type max_edge: int
    :param min_edge: minimum number of edges (connected to each vertex)
    :type min_edge: int
    :param min_weight: weight min range
    :type min_weight: int
    :param max_weight: weight max range
    :type max_weight: int
    :param precision: numbers precision
    :type precision: int
    :param sign: weight sign flag
    :type sign: bool
    :param direct: directed and undirected graph flag
    :type direct: bool
    :param self_loop: self loop flag
    :type self_loop: bool
    :param multigraph: multigraph flag
    :type multigraph: bool
    :param used_vertices: used vertices dictionary
    :type used_vertices: dict
    :param degree_dict: all vertices degree
    :type degree_dict: dict
    :param degree_sort_dict: degree to vertices list
    :type degree_sort_dict: dict
    :return: branch and weight list
    """
    index = 0
    branch_list = []
    weight_list = []
    reference_vertices = []
    random_unit = randint
    vertex_degree = degree_dict[vertex_index]
    if vertex_degree >= max_edge:
        return [branch_list, weight_list]
    threshold = threshold_calc(
        min_edge=min_edge,
        max_edge=max_edge,
        vertex_degree=vertex_degree)
    for i in range(max_edge + 1):
        reference_vertices.extend(list(degree_sort_dict[i].values()))
        if len(reference_vertices) >= threshold:
            break
    if precision > 0:
        random_unit = uniform
    if not direct and (
            vertex_index in used_vertices) and not multigraph:
        reference_vertices = list(
            set(reference_vertices) - set(used_vertices[vertex_index]))
    if not self_loop and vertex_index in reference_vertices:
        reference_vertices.remove(vertex_index)
    if int(os.environ.get("PYRGG_TEST_MODE", 0)):
        reference_vertices.sort()
    while (index < threshold):
        vertex_degree = degree_dict[vertex_index]
        if vertex_degree >= max_edge:
            break
        if len(reference_vertices) == 0:
            break
        random_tail_index = choice(
            range(len(reference_vertices)))
        random_tail = reference_vertices[random_tail_index]
        random_tail_degree = degree_dict[random_tail]
        if random_tail_degree >= max_edge or (
            random_tail == vertex_index and random_tail_degree >= (
                max_edge - 1)):
            reference_vertices.pop(random_tail_index)
            continue
        if not direct:
            try:
                used_vertices[random_tail].append(vertex_index)
            except KeyError:
                used_vertices[random_tail] = [vertex_index]
        weight_sign = 1
        if sign:
            weight_sign = choice([1, -1])
        random_weight = weight_sign * random_unit(min_weight, max_weight)
        random_weight = round(random_weight, precision)
        branch_list.append(random_tail)
        weight_list.append(random_weight)
        index += 1
        del degree_sort_dict[vertex_degree][vertex_index]
        degree_dict[random_tail] += 1
        degree_dict[vertex_index] += 1
        degree_sort_dict[degree_dict[vertex_index]
                         ][vertex_index] = vertex_index
        if random_tail != vertex_index:
            del degree_sort_dict[random_tail_degree][random_tail]
            degree_sort_dict[degree_dict[random_tail]
                             ][random_tail] = random_tail
        if not multigraph:
            reference_vertices.pop(random_tail_index)
    return [branch_list, weight_list]


def edge_gen(
        vertices_number,
        min_weight,
        max_weight,
        min_edge,
        max_edge,
        sign,
        direct,
        self_loop,
        multigraph):
    """
    Generate each vertex connection number.

    :param vertices_number: number of vertices
    :type vertices_number: int
    :param min_weight: weight min range
    :type min_weight: int
    :param max_weight: weight max range
    :type max_weight: int
    :param min_edge: minimum number of edges (connected to each vertex)
    :type min_edge: int
    :param max_edge: maximum number of edges (connected to each vertex)
    :type max_edge: int
    :param sign: weight sign flag
    :type sign: bool
    :param direct: directed and undirected graph flag
    :type direct: bool
    :param self_loop: self loop flag
    :type self_loop: bool
    :param multigraph: multigraph flag
    :type multigraph: bool
    :return: list of dicts
    """
    precision = max(
        get_precision(max_weight),
        get_precision(min_weight))
    temp = 0
    vertices_id = list(range(1, vertices_number + 1))
    vertices_edge = []
    weight_list = []
    used_vertices = {}
    degree_sort_dict = {i: {} for i in range(max_edge + 1)}
    degree_dict = {}
    for i in vertices_id:
        degree_dict[i] = 0
        degree_sort_dict[0][i] = i
    branch_gen_params = {
        "max_edge": max_edge,
        "min_edge": min_edge,
        "min_weight": min_weight,
        "max_weight": max_weight,
        "sign": sign,
        "direct": direct,
        "self_loop": self_loop,
        "multigraph": multigraph,
        "used_vertices": used_vertices,
        "degree_dict": degree_dict,
        "degree_sort_dict": degree_sort_dict,
        "precision": precision}
    for i in vertices_id:
        temp_list = branch_gen(vertex_index=i, **branch_gen_params)
        vertices_edge.append(temp_list[0])
        weight_list.append(temp_list[1])
        temp = temp + len(temp_list[0])
    return [dict(zip(vertices_id, vertices_edge)),
            dict(zip(vertices_id, weight_list)), temp]


def gen_using(
        gen_function,
        file_name,
        input_dict):
    """
    Generate graph using given function based on PyRGG model.

    :param gen_function: generation function
    :type gen_function: function object
    :param file_name: file name
    :type file_name: str
    :param input_dict: input data
    :type input_dict: dict
    :return: number of edges as int
    """
    edge_dic, weight_dic, edge_number = edge_gen(
        input_dict['vertices'],
        input_dict['min_weight'],
        input_dict['max_weight'],
        input_dict['min_edge'],
        input_dict['max_edge'],
        input_dict['sign'],
        input_dict['direct'],
        input_dict['self_loop'],
        input_dict['multigraph'])
    min_weight, max_weight = get_min_max_weight(weight_dic)
    weighted = is_weighted(max_weight, min_weight, bool(input_dict['sign']))
    gen_function(
        edge_dic,
        weight_dic,
        {
            "file_name": file_name,
            "vertices_number": input_dict['vertices'],
            "weighted": weighted,
            "edge_number": edge_number,
            "min_weight": min_weight,
            "max_weight": max_weight,
            "direct": input_dict['direct'],
            "multigraph": is_multigraph(edge_dic),
        })
    return edge_number


def logger(file, file_name, elapsed_time, input_dict):
    """
    Save generated graph logs for PyRGG engine.

    :param file: file to write log into
    :type file: file object
    :param file_name: file name
    :type file_name: str
    :param elapsed_time: elapsed time
    :type elapsed_time: str
    :param input_dict: input data
    :type input_dict: dict
    :return: None
    """
    try:
        text = "Vertices : {0}\n".format(input_dict['vertices'])
        text += "Total Edges : {0}\n".format(input_dict['edge_number'])
        text += "Max Edge : {0}\n".format(input_dict['max_edge'])
        text += "Min Edge : {0}\n".format(input_dict['min_edge'])
        text += "Directed : {0}\n".format(bool(input_dict['direct']))
        text += "Signed : {0}\n".format(bool(input_dict['sign']))
        text += "Multigraph : {0}\n".format(bool(input_dict['multigraph']))
        text += "Self Loop : {0}\n".format(bool(input_dict['self_loop']))
        text += "Weighted : {0}\n".format(
            is_weighted(input_dict['max_weight'], input_dict['min_weight'], bool(input_dict['sign'])))
        text += "Max Weight : {0}\n".format(input_dict['max_weight'])
        text += "Min Weight : {0}\n".format(input_dict['min_weight'])
        text += "Engine : {0} ({1})\n".format(
            input_dict['engine'], ENGINE_MENU[input_dict['engine']])
        save_log(file, file_name, elapsed_time, text)
    except Exception:
        print(PYRGG_LOGGER_ERROR_MESSAGE)



# -*- coding: utf-8 -*-
"""Stochastic Block Model Engine module."""
from random import random
from itertools import combinations
from pyrgg.params import ENGINE_MENU, PYRGG_LOGGER_ERROR_MESSAGE
from pyrgg.functions import save_log


def edge_gen(
        vertices,
        block_sizes,
        probability_matrix,
        direct,
        self_loop):
    """
    Generate each vertex connection number.

    :param vertices: number of vertices
    :type vertices: int
    :param block_sizes: block sizes
    :type block_sizes: list
    :param probability_matrix: probability matrix
    :type probability_matrix: list
    :param direct: directed graph flag
    :type direct: bool
    :param self_loop: self loop flag
    :type self_loop: bool
    :return: list of dicts
    """
    edge_number = 0
    edge_dic = {x: [] for x in range(1, vertices + 1)}
    weight_dic = {x: [] for x in range(1, vertices + 1)}
    vertices2block = {}
    for c, r in enumerate(block_sizes):
        n0 = len(vertices2block)
        vertices2block.update({i: c for i in range(n0 + 1, n0 + r + 1)})
    
    vertices_pairs = list(combinations(range(1, vertices + 1), 2))
    if direct:
        vertices_pairs += [(j, i) for i, j in vertices_pairs]
    if self_loop:
        vertices_pairs += [(i, i) for i in range(1, vertices + 1)]
    for v1, v2 in sorted(vertices_pairs):
        c1 = vertices2block[v1]
        c2 = vertices2block[v2]
        if random() < probability_matrix[c1][c2]:
            edge_dic[v1].append(v2)
            weight_dic[v1].append(1)
            edge_number += 1
    return [edge_dic, weight_dic, edge_number]
 

def gen_using(
        gen_function,
        file_name,
        input_dict):
    """
    Generate graph using given function based on Stochastic Block Model model.

    Refer to (https://en.wikipedia.org/wiki/Stochastic_block_model).

    :param gen_function: generation function
    :type gen_function: function object
    :param file_name: file name
    :type file_name: str
    :param input_dict: input data
    :type input_dict: dict
    :return: number of edges as int
    """
    edge_dic, weight_dic, edge_number = edge_gen(
        input_dict['vertices'],
        input_dict['block_sizes'],
        input_dict['probability_matrix'],
        input_dict['direct'],
        input_dict['self_loop'])
    gen_function(
        edge_dic,
        weight_dic,
        {
            "file_name": file_name,
            "vertices_number": input_dict['vertices'],
            "edge_number": edge_number,
            "weighted": False,
            "max_weight": 1,
            "min_weight": 1,
            "direct": input_dict['direct'],
            "multigraph": False,
        })
    return edge_number


def logger(file, file_name, elapsed_time, input_dict):
    """
    Save generated graph logs for Stochastic Block Model engine.

    :param file: file to write log into
    :type file: file object
    :param file_name: file name
    :type file_name: str
    :param elapsed_time: elapsed time
    :type elapsed_time: str
    :param input_dict: input data
    :type input_dict: dict
    :return: None
    """
    try:
        text = "Vertices : {0}\n".format(input_dict['vertices'])
        text += "Total Edges : {0}\n".format(input_dict['edge_number'])
        text += "Block Sizes : {0}\n".format(input_dict['block_sizes'])
        text += "Probability Matrix : {0}\n".format(input_dict['probability_matrix'])
        text += "Directed : {0}\n".format(bool(input_dict['direct']))
        text += "Self Loop : {0}\n".format(bool(input_dict['self_loop']))
        text += "Engine : {0} ({1})\n".format(input_dict['engine'], ENGINE_MENU[input_dict['engine']])
        save_log(file, file_name, elapsed_time, text)
    except Exception:
        print(PYRGG_LOGGER_ERROR_MESSAGE)



# -*- coding: utf-8 -*-
"""Erdős-Rényi Engine module."""
from random import shuffle
from pyrgg.params import ENGINE_MENU, PYRGG_LOGGER_ERROR_MESSAGE
from pyrgg.functions import save_log


def edge_gen(n, m, direct):
    """
    Generate each vertex connection number.

    :param n: number of vertices
    :type n: int
    :param m: number of edges
    :type m: int
    :param direct: directed graph flag
    :type direct: bool
    :return: list of dicts
    """
    edge_dic = {}
    weight_list = []
    edge_mold = []
    max_edge = (n * (n - 1)) // 2
    if direct:
        max_edge *= 2
    m = min(m, max_edge)
    edge_mold = m * [1] + (max_edge - m) * [0]
    shuffle(edge_mold)
    for i in range(1, n + 1):
        edge_dic[i] = []
        temp_list = []
        dest_list = range(i + 1, n + 1)
        if direct:
            dest_list = [*range(1, i), *dest_list]
        for j in dest_list:
            if edge_mold.pop() == 1:
                temp_list.append(1)
                edge_dic[i].append(j)
        weight_list.append(temp_list)
    return [edge_dic, dict(zip(range(1, n + 1), weight_list)), m]


def gen_using(
        gen_function,
        file_name,
        input_dict):
    """
    Generate graph using given function based on Erdos Renyi - G(n, m) model.

    Refer to (https://en.wikipedia.org/wiki/Erd%C5%91s%E2%80%93R%C3%A9nyi_model).

    :param gen_function: generation function
    :type gen_function: function object
    :param file_name: file name
    :type file_name: str
    :param input_dict: input data
    :type input_dict: dict
    :return: number of edges as int
    """
    edge_dic, weight_dic, edge_number = edge_gen(
        input_dict['vertices'],
        input_dict['edge_number'],
        input_dict['direct'])
    gen_function(
        edge_dic,
        weight_dic,
        {
            "file_name": file_name,
            "vertices_number": input_dict['vertices'],
            "edge_number": edge_number,
            "weighted": False,
            "max_weight": 1,
            "min_weight": 1,
            "direct": input_dict['direct'],
            "multigraph": False,
        })
    return edge_number


def logger(file, file_name, elapsed_time, input_dict):
    """
    Save generated graph logs for Erdős-Rényi engine.

    :param file: file to write log into
    :type file: file object
    :param file_name: file name
    :type file_name: str
    :param elapsed_time: elapsed time
    :type elapsed_time: str
    :param input_dict: input data
    :type input_dict: dict
    :return: None
    """
    try:
        text = "Vertices : {0}\n".format(input_dict['vertices'])
        text += "Total Edges : {0}\n".format(input_dict['edge_number'])
        text += "Directed : {0}\n".format(bool(input_dict['direct']))
        text += "Engine : {0} ({1})\n".format(input_dict['engine'], ENGINE_MENU[input_dict['engine']])
        save_log(file, file_name, elapsed_time, text)
    except Exception:
        print(PYRGG_LOGGER_ERROR_MESSAGE)



# -*- coding: utf-8 -*-
"""Setup module."""
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
from textwrap import fill

_description = """\
PyRGG is a user-friendly synthetic random graph generator that is written in Python and supports multiple graph file formats, such as DIMACS-Graph files.
It can generate graphs of various sizes and is specifically designed to create input files for a wide range of graph-based research applications, including testing,
benchmarking, and performance analysis of graph processing frameworks.
PyRGG is aimed at computer scientists who are studying graph algorithms and graph processing frameworks.
"""

PYRGG_DESCRIPTION = fill(_description, width=113)


def get_requires():
    """Read requirements.txt."""
    requirements = open("requirements.txt", "r").read()
    return list(filter(lambda x: x != "", requirements.split()))


def read_description():
    """Read README.md and CHANGELOG.md."""
    try:
        with open("README.md") as r:
            description = "\n"
            description += r.read()
        with open("CHANGELOG.md") as c:
            description += "\n"
            description += c.read()
        return description
    except Exception:
        return PYRGG_DESCRIPTION.replace("\n", " ")


setup(
    name='pyrgg',
    packages=['pyrgg', 'pyrgg.engines'],
    version='1.7',
    description='Python Random Graph Generator',
    long_description=read_description(),
    long_description_content_type='text/markdown',
    author='PyRGG Development Team',
    author_email='info@pyrgg.site',
    url='https://github.com/sepandhaghighi/pyrgg',
    download_url='https://github.com/sepandhaghighi/pyrgg/tarball/v1.7',
    keywords='random graph python3 python generator graph-process generator DIMACS JSON YAML Pickle CSV TSV WEL ASP TGF UCINET',
    project_urls={
        'Webpage': 'https://www.pyrgg.site',
        'Source': 'https://github.com/sepandhaghighi/pyrgg',
    },
    install_requires=get_requires(),
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Manufacturing',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Education',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Human Machine Interfaces',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Physics',
    ],
    license='MIT',
    entry_points={
        'console_scripts': [
            'pyrgg = pyrgg.__main__:main',
        ]})
